#include <mpi.h>
#include <vector>
#include "defs.h"
#include "alloc.h"
#include "timer.h"
#include "mpimacros.h"

void peer2peer_exchange(const std::vector<int> &params, olocal &output) {
    if (params.size() != 6) {
        throw std::runtime_error("peer2peer_exchange: wrong parameters");
    }
    int rank = params[0];
    int size = params[1];
    int distance = params[2];
    int npeers_per_node = params[3];
    bool use_gpu_memory = (bool)params[4];
    bool bidirectional = (bool)params[5];
    bool is_intra_node = (2 * distance <= npeers_per_node) && (npeers_per_node % distance == 0);
    std::string kind = (is_intra_node ? "intra-node" : "cross-node");
    int r = rank % (2 * distance);
    bool do_recv_first = false;
    int peer = rank + distance;
    if (r >= distance) {
        do_recv_first = true;
        peer = rank - distance; 
    } 
    MPI_Comm comm;
	if (is_intra_node) {
#if WITH_PER_NODE_COMM
	    MPI_CALL(MPI_Comm_split_type(MPI_COMM_WORLD, MPI_COMM_TYPE_SHARED, 0, MPI_INFO_NULL, &comm));
        peer = peer % npeers_per_node;
#else
		MPI_CALL(MPI_Comm_dup(MPI_COMM_WORLD, &comm));
#endif
	} else {
		MPI_CALL(MPI_Comm_dup(MPI_COMM_WORLD, &comm));
	}
    if (peer >= size || peer < 0) {
	    MPI_CALL(MPI_Comm_free(&comm));
        return;  
    }
    int niters = 10;
    int nwarmup = 3;
    size_t bufsize = (1<<26) * sizeof(real);
    size_t cachesize = 1<<23;
    char *buf = nullptr;
    if (use_gpu_memory) {
        buf = sys::gpu_alloc(bufsize + cachesize);
    } else {
#if PINNED_FOR_MPI_TRANSFERS
        buf = sys::cpu_alloc_pinned(bufsize + cachesize);
#else
        buf = sys::cpu_alloc_pageable(bufsize + cachesize);
#endif
    }
    std::vector<double> result;
	if (!bidirectional) {
        timer t(rank2str(rank) + "Peer-to-peer bandwidth unidirectional:", bufsize * 2, niters, result);
        t.comment << kind << " with distance=" << distance;
        for (int i = 0; i < niters + nwarmup; i++) {
            if (i == nwarmup)
                t.start();
            if (do_recv_first) {
                MPI_CALL(MPI_Recv(buf + (i%2) * cachesize, bufsize, MPI_BYTE, peer, 0, comm, MPI_STATUS_IGNORE));
                MPI_CALL(MPI_Send(buf + (i%2) * cachesize, bufsize, MPI_BYTE, peer, 0, comm));
            } else {
                MPI_CALL(MPI_Send(buf + (i%2) * cachesize, bufsize, MPI_BYTE, peer, 0, comm));
                MPI_CALL(MPI_Recv(buf + (i%2) * cachesize, bufsize, MPI_BYTE, peer, 0, comm, MPI_STATUS_IGNORE));
            }
        }
    } else {
        timer t(rank2str(rank) + "Peer-to-peer bandwidth bidirectional:", bufsize * 2, niters, result);
        t.comment << kind << " with distance=" << distance;
		MPI_Request request;
        for (int i = 0; i < niters + nwarmup; i++) {
            if (i == nwarmup)
                t.start();
			MPI_CALL(MPI_Isend(buf + (i%2) * cachesize, bufsize, MPI_BYTE, peer, 0, comm, &request));
			MPI_CALL(MPI_Recv(buf + (i%2) * cachesize, bufsize, MPI_BYTE, peer, 0, comm, MPI_STATUS_IGNORE));
			MPI_CALL(MPI_Wait(&request, MPI_STATUS_IGNORE));
        }
    }
    if (use_gpu_memory) {
        sys::gpu_free(buf);
    } else {
#if PINNED_FOR_MPI_TRANSFERS
        sys::cpu_free_pinned(buf);
#else
        sys::cpu_free_pageable(buf);
#endif
    }
	MPI_CALL(MPI_Comm_free(&comm));
    output.values.push_back(result);
    output.header = std::string("MPI_P2P_") + (is_intra_node ? "intranode_" : "crossnode_") + (use_gpu_memory ? "GPU_" : "CPU_") + (bidirectional ? "1DIR" : "2DIR");
}

